from MAPI import *
from MAPI.Util import *
from MAPI.Defs import *

ROLE_OWNER = 0x5fb
ROLE_REVIEWER = 0x401

def create(folder, flags = 0):
    return ACL(folder, flags)

class ACL:
    def __init__(self, folder, flags = 0):
        self.folder = folder
        self.flags = flags
        self.acls = self.folder.OpenProperty(PR_ACL_TABLE, IID_IExchangeModifyTable, 0, self.flags & (MAPI_MODIFY | MAPI_BEST_ACCESS))

    def __GetRightsByID(self, userid):
        row = self.__GetACLEntry(SPropertyRestriction(RELOP_EQ, PR_MEMBER_ENTRYID, SPropValue(PR_MEMBER_ENTRYID, userid)), [PR_MEMBER_RIGHTS])
        if PROP_TYPE(row[0].ulPropTag) == PT_ERROR:
            raise MAPIError.from_hresult(row[0].Value)
        return row[0].Value

    def GetRightsByName(self, username):
        row = self.__GetACLEntry(SPropertyRestriction(RELOP_EQ, PR_MEMBER_NAME, SPropValue(PR_MEMBER_NAME, username)), [PR_MEMBER_RIGHTS])
        if PROP_TYPE(row[0].ulPropTag) == PT_ERROR:
            raise MAPIError.from_hresult(row[0].Value)
        return row[0].Value

    def __GetACLEntry(self, restriction, columns):
        table = self.acls.GetTable(0)
        table.SetColumns(columns, TBL_BATCH)
        table.FindRow(restriction, BOOKMARK_BEGINNING, 0)
        return table.QueryRows(1, 0)[0] # Must return one because FindRow would have thrown a MAPIError if the entry was not found

    def UpdateFrom(self, other, **kwargs):
        rowlist = []        
        mask = None
        owner = None
        hide_all = False
        if 'mask' in kwargs: mask = kwargs['mask']
        if 'owner' in kwargs: owner = kwargs['owner']
        if 'hide_all' in kwargs: hide_all = kwargs['hide_all']

        if owner and owner.acl_capable:
            # See if the owner has an entry. If so, just add that as is.
            try:
                row = self.__GetACLEntry(SPropertyRestriction(RELOP_EQ, PR_MEMBER_NAME, SPropValue(PR_MEMBER_NAME, owner.name)), [PR_MEMBER_ENTRYID, PR_MEMBER_RIGHTS])
                rowlist.append(ROWENTRY(ROW_ADD, row))
            except MAPIErrorNotFound: pass

        table = other.acls.GetTable(0)
        table.SetColumns([PR_MEMBER_ENTRYID, PR_MEMBER_RIGHTS], TBL_BATCH)
        if owner and owner.acl_capable:
            table.Restrict(SPropertyRestriction(RELOP_NE, PR_MEMBER_NAME, SPropValue(PR_MEMBER_NAME, owner.name)), TBL_BATCH)

        if hide_all:
            everyone_id = '00000000ac21a95040d3ee48b319fba75330442500000000080000000100000000000000'.decode('hex')
            rowlist.append(ROWENTRY(ROW_ADD, [SPropValue(PR_MEMBER_RIGHTS, 0), SPropValue(PR_MEMBER_ENTRYID, everyone_id)]))

        while True:
            rows = table.QueryRows(32, 0)
            if len(rows) == 0:
                break

            for row in rows:
                if not mask is None:
                    row[1].Value = row[1].Value & mask
                rowlist.append(ROWENTRY(ROW_ADD, row))

        self.acls.ModifyTable(ROWLIST_REPLACE, rowlist)

    def AddFolderViewFrom(self, other):
        rowlist = []

        table = self.acls.GetTable(0)
        table.SetColumns([PR_MEMBER_ID, PR_MEMBER_RIGHTS], TBL_BATCH)

        other_table = other.acls.GetTable(0)
        other_table.SetColumns([PR_MEMBER_ENTRYID, PR_MEMBER_RIGHTS], TBL_BATCH)

        while True:
            rows = other_table.QueryRows(32, 0)
            if len(rows) == 0:
                break

            for row in rows:
                try:
                    table.FindRow(SPropertyRestriction(RELOP_EQ, PR_MEMBER_ENTRYID, SPropValue(PR_MEMBER_ENTRYID, row[0].Value)), BOOKMARK_BEGINNING, 0)
                    result = table.QueryRows(1, 0)[0]
                    # Only update if the current rights doesn't include folder view
                    if result[1].Value & 0x400 == 0:
                        result[1].Value |= 0x400
                        rowlist.append(ROWENTRY(ROW_MODIFY, result))
                except MAPIErrorNotFound:
                    rowlist.append(ROWENTRY(ROW_ADD, [row[0], SPropValue(PR_MEMBER_RIGHTS, 0x400)]))

        self.acls.ModifyTable(0, rowlist)
    
    def SetRights(self, owner, writable):
        if not owner.acl_capable:
            return False
        rowlist = []
        rights = {True: ROLE_OWNER,
                  False: ROLE_REVIEWER}.get(writable)
        new_row = [SPropValue(PR_MEMBER_ENTRYID, owner.entryid), SPropValue(PR_MEMBER_RIGHTS, rights)]
        try:
            row = self.__GetACLEntry(SPropertyRestriction(RELOP_EQ, PR_MEMBER_NAME, SPropValue(PR_MEMBER_NAME, owner.name)), [PR_MEMBER_RIGHTS])
        except MAPIErrorNotFound:
            rowlist.append(ROWENTRY(ROW_ADD, new_row))
        else:
            if row[0].Value != rights:
                rowlist.append(ROWENTRY(ROW_MODIFY, new_row))
        if rowlist:
            self.acls.ModifyTable(0, rowlist)
        return bool(rowlist)
