import random

from helpers import StoreHelper, FolderHelper
import acl

from MAPI import *
from MAPI.Tags import *
from MAPI.Struct import *
from MAPI.Util import *
from MAPI.Util.Generators import GetStores
from MAPI.Util.Iterators import FolderIterator

import licensereq

class Stats:
    class TypedStats:
        def __init__(self):
            self.success = 0
            self.failed = 0
            self.skipped = 0
    
    def __init__(self):
        self.storestats = self.TypedStats()
        self.folderstats = self.TypedStats()
        self.archivestats = self.TypedStats()
        self.accessstats = self.TypedStats()

    def __getattr__(self, attr):
        if attr == 'retval':
            if self.storestats.failed == 0 and self.folderstats.failed == 0 and self.archivestats.failed == 0 and self.accessstats.failed == 0:
                return 0
            return 1
        else:
            raise AttributeError()

    def __str__(self):
        str =  "Successfully processed %u store(s).\n" % self.storestats.success
        str += "Skipped %u unarchived store(s).\n" % self.storestats.skipped
        str += "%u store(s) failed.\n" % self.storestats.failed
        str += "Successfully processed %u folder(s).\n" % self.folderstats.success
        str += "Skipped %u unarchived folder(s), not counting folders of skipped stores.\n" % self.folderstats.skipped
        str += "%u folder(s) failed.\n" % self.folderstats.failed
        str += "Successfully updated ACL table of %u archive folder(s).\n" % self.archivestats.success
        str += "Skipped %u missing archive folder(s).\n" % self.archivestats.skipped
        str += "Failed to update ACL table of %u archive folder(s).\n" % self.archivestats.failed
        str += "Successfully updated access rights of %u archive(s).\n" % self.accessstats.success
        str += "Skipped %u missing archive root(s).\n" % self.accessstats.skipped
        str += "Failed to update access rights of %u archive(s).\n" % self.accessstats.failed
        return str



class ACLSyncer:
    def __init__(self, storehelper):
        self.storehelper = storehelper
        self.success = True

    def __GetIpmSubtree(self, store):
        id = HrGetOneProp(store, PR_IPM_SUBTREE_ENTRYID).Value
        return store.OpenEntry(id, IID_IMAPIFolder, MAPI_MODIFY)

    def __IsIpmSubtree(self, store, folder):
        ipmid = HrGetOneProp(store, PR_IPM_SUBTREE_ENTRYID).Value
        fldid = HrGetOneProp(folder, PR_ENTRYID).Value
        return store.CompareEntryIDs(ipmid, fldid, 0)        

    def UpdateStoreAccess(self):
        '''How this works:
           A user will set permissions through the permissions tab on the store.
           The archiver will grant folder view permissions on the archive IPM Subtree
           and read-only (or write) on the root folder for the users archive. If that
           turns out to be the IPM Subtree, these settings will overrule the folder
           view permissions.
           So what we'll do here is grant folder view permissions for all users that
           have any permissions on the users store on the IPM Subtree of the archive
           and grant as much permissions on the root of the users archive. SO that will
           be the combination of the permissions granted on the users store and the
           permissions the user has himself on the archive root.'''
        global stats        
        for archive in self.storehelper.GetArchives():
            if archive is None:
                stats.accessstats.skipped += 1
                continue

            try:
                store_acl = acl.create(self.storehelper.store)
                archive_acl = acl.create(archive[1])

                if self.storehelper.owner.acl_capable:
                    owner_rights = archive_acl.GetRightsByName(self.storehelper.owner.name)
                    archive_acl.UpdateFrom(store_acl, owner = self.storehelper.owner, mask = owner_rights, hide_all = True)
                else:
                    archive_acl.UpdateFrom(store_acl, owner = self.storehelper.owner, mask = acl.ROLE_REVIEWER, hide_all = True)

                if not self.__IsIpmSubtree(archive[0], archive[1]):
                    archive_ipm = self.__GetIpmSubtree(archive[0])
                    archive_ipm_acl = acl.create(archive_ipm)
                    archive_ipm_acl.AddFolderViewFrom(store_acl)
                stats.accessstats.success += 1
            except MAPIError, err:
                print "Unhandled MAPI error occurred while updating archive access, hr=0x%08x." % err.hr
                stats.accessstats.failed += 1
                self.success = False
            except:
                print "Unknown error occurred while updating archive access."
                stats.accessstats.failed += 1
                self.success = False

    def ProcessFolder(self, folder, depth):
        global stats

        # The IPM Subtree is handled in UpdateStoreAccess (where the rights are actually obtained from the store.
        if self.__IsIpmSubtree(self.storehelper.store, folder):
            return
            
        try:
            folderhelper = FolderHelper(self.storehelper, folder)
            if not folderhelper.IsArchived():
                stats.folderstats.skipped += 1
                return

            archives = folderhelper.GetArchives()
            
            # If the store for a folder failed to open, None will be returned in the list
            for archive in archives:
                if archive is None:
                    stats.archivestats.skipped += 1
                    continue

                try:
                    self.ProcessArchive(archive, folderhelper)
                    stats.archivestats.success += 1
                except MAPIError, err:
                    print "Unhandled MAPI error occurred while updating archive, hr=0x%08x." % err.hr
                    stats.archivestats.failed += 1
                    self.success = False
                except:
                    print "Unknown error occurred while updating archive."
                    stats.archivestats.failed += 1
                    self.success = False

            if self.success:
                stats.folderstats.success += 1
            else:
                stats.folderstats.failed += 1
        except MAPIError, err:
            print "Unhandled MAPI error occurred while processing folder, hr=0x%08x." % err.hr
            stats.folderstats.failed += 1
            self.success = False
        except:
            print "Unknown error occurred while processing folder."
            stats.folderstats.failed += 1
            self.success = False

    def ProcessArchive(self, archive, folderhelper):
        source_acl = acl.create(folderhelper.folder)
        archive_acl = acl.create(archive[1], MAPI_MODIFY)

        if folderhelper.owner.acl_capable:
            owner_rights = folderhelper.GetOwnerArchiveRights(archive[1])
            archive_acl.UpdateFrom(source_acl, owner = folderhelper.owner, mask = owner_rights)
        else:
            archive_acl.UpdateFrom(source_acl, owner = folderhelper.owner, mask = acl.ROLE_REVIEWER)


def main(session, users):
    r = random.randint(0, 0xffffffff)
    request = licensereq.CreateLicenseRequest(r, "ARCHIVER", 0, None)
    license = GetDefaultStore(session).QueryInterface(IID_IECLicense)
    response = license.LicenseAuth(request)
    try:
        licensereq.ProcessLicenseResponse(r, response)
    except RuntimeError:
        print "No valid archiver license found."
        return 2

    global stats
    stats = Stats()
    
    if len(users) == 0:
        users = None
    for store in GetStores(session, users, 0):
        helper = StoreHelper(session, store)
        if not helper.IsArchived():
            stats.storestats.skipped += 1
            continue

        syncer = ACLSyncer(helper)
        syncer.UpdateStoreAccess()
        
        fi = FolderIterator(store, depth=0)
        fi.Iterate(syncer.ProcessFolder)
        
        if syncer.success:
            stats.storestats.success += 1
        else:
            stats.storestats.failed += 1

    print '**** ACL Sync results ****'
    print stats
    return stats.retval
