from MAPI import *
from MAPI.Util import *
from MAPI.Defs import *

import acl

PSETID_Archive = DEFINE_GUID(0x72e98ebc, 0x57d2, 0x4ab5, 0xb0, 0xaa, 0xd5, 0x0a, 0x7b, 0x53, 0x1c, 0xb9)

class _BaseHelper:
    def __init__(self, object):
        self.object = object
        
        names = [MAPINAMEID(PSETID_Archive, MNID_STRING, u'store-entryids'),
                 MAPINAMEID(PSETID_Archive, MNID_STRING, u'item-entryids')]
        ids = object.GetIDsFromNames(names, MAPI_CREATE)
        self.PROP_STORE_ENTRYIDS = CHANGE_PROP_TYPE(ids[0], PT_MV_BINARY)
        self.PROP_ITEM_ENTRYIDS = CHANGE_PROP_TYPE(ids[1], PT_MV_BINARY)

    def IsArchived(self):
        props = self.object.GetProps([self.PROP_STORE_ENTRYIDS, self.PROP_ITEM_ENTRYIDS], 0)
        return PROP_TYPE(props[0].ulPropTag) != PT_ERROR and PROP_TYPE(props[1].ulPropTag) != PT_ERROR and len(props[0].Value) == len(props[1].Value)


class StoreOwner(object):
    def __init__(self, session, store):
        self.entryid = HrGetOneProp(store, PR_MAILBOX_OWNER_ENTRYID).Value
        self._owner = session.OpenEntry(self.entryid, None, 0)
        self.name, display_type_ex = (prop.Value for prop in self._owner.GetProps([PR_DISPLAY_NAME, PR_DISPLAY_TYPE_EX], 0))
        self.acl_capable = (display_type_ex & DTE_FLAG_ACL_CAPABLE != 0)


class StoreHelper(_BaseHelper):
    def __init__(self, session, store):
        _BaseHelper.__init__(self, store)
        self.session = session
        self.store = store
        self.owner = StoreOwner(session, store)
        self.storecache = {}
        self.rootcache = {}

    def GetMsgStore(self, storeid):
        if storeid in self.storecache:
            return self.storecache[storeid]
        store = self.session.OpenMsgStore(0, storeid, IID_IMsgStore, MDB_WRITE)
        self.storecache[storeid] = store
        return store

    def GetArchives(self):
        props = self.object.GetProps([self.PROP_STORE_ENTRYIDS, self.PROP_ITEM_ENTRYIDS], 0)
        return [self.__OpenEntry(entry) for entry in zip(props[0].Value, props[1].Value)]

    def GetArchiveStores(self):
        ids = HrGetOneProp(self.object, self.PROP_STORE_ENTRYIDS)
        return [self.GetMsgStore(id) for id in ids.Value]

    def GetArchiveRoot(self, storeid):
        if storeid in self.rootcache:
            return self.rootcache[storeid]
        props = self.store.GetProps([self.PROP_STORE_ENTRYIDS, self.PROP_ITEM_ENTRYIDS], 0)
        for entry in zip(props[0].Value, props[1].Value):
            result = self.session.CompareEntryIDs(entry[0], storeid, 0)
            if result != 0:
                store = self.GetMsgStore(storeid)
                try: root = store.OpenEntry(entry[1], IID_IMAPIFolder, MAPI_MODIFY)
                except MAPIError:
                    continue
                self.rootcache[storeid] = root
                return root
        raise MAPIErrorNotFound()


    def _DetachAllArchives(self):
        try:
            prop = HrGetOneProp(self.object, self.PROP_STORE_ENTRYIDS)
        except MAPIErrorNotFound:
            return []
        self.object.DeleteProps([self.PROP_STORE_ENTRYIDS, self.PROP_ITEM_ENTRYIDS])
        return prop.Value


    def _DetachArchives(self, archiveIds):
        props = self.object.GetProps([self.PROP_STORE_ENTRYIDS, self.PROP_ITEM_ENTRYIDS], 0)
        storeEntryIDs = SPropValue(self.PROP_STORE_ENTRYIDS, [])
        itemEntryIDs = SPropValue(self.PROP_ITEM_ENTRYIDS, [])
        for archive in zip(props[0].Value, props[1].Value):
            if not archive[0] in archiveIds:
                storeEntryIDs.Value.append(archive[0])
                itemEntryIDs.Value.append(archive[1])
        self.object.SetProps([storeEntryIDs, itemEntryIDs])


    def DetachArchives(self, archiveIds):
        if archiveIds is None:
            archiveIds = self._DetachAllArchives()
        else:
            self._DetachArchives(archiveIds)

        removableStores = []
        for archiveId in archiveIds:
            try:
                archive = self.GetMsgStore(archiveId)
            except MAPIErrorNotFound:
                continue
            # The preferred way is to get the PR_EC_STORETYPE property. However that's not
            # available on all servers yet. So go for the PR_MDB_PROVIDER instead.
            mdbProvider = HrGetOneProp(archive, PR_MDB_PROVIDER).Value
            if mdbProvider == ZARAFA_STORE_ARCHIVE_GUID:
                removableStores.append(archive)

        return removableStores


    def __OpenEntry(self, entry):
        try:
            store = self.GetMsgStore(entry[0])
            return (store, store.OpenEntry(entry[1], IID_IMAPIFolder, MAPI_MODIFY))
        except MAPIError:
            return None



class FolderHelper(_BaseHelper):
    def __init__(self, storehelper, folder):
        _BaseHelper.__init__(self, folder)
        self.storehelper = storehelper
        self.folder = folder
        self.owner = storehelper.owner

    def GetArchives(self):
        props = self.object.GetProps([self.PROP_STORE_ENTRYIDS, self.PROP_ITEM_ENTRYIDS], 0)
        if props[0].ulPropTag == self.PROP_STORE_ENTRYIDS or props[1].ulPropTag == self.PROP_ITEM_ENTRYIDS:
            return [self.__OpenEntry(entry) for entry in zip(props[0].Value, props[1].Value)]
        return []

    def GetArchiveRoot(self, folder):
        storeid = folder.GetProps([PR_STORE_ENTRYID], 0)[0].Value
        return self.storehelper.GetArchiveRoot(storeid)

    def GetOwnerArchiveRights(self, archive):
        if self.owner.acl_capable:
            try:
                aclFolder = acl.create(archive)
                return aclFolder.GetRightsByName(self.owner.name)
            except MAPIErrorNotFound:
                pass
            try:
                root = self.GetArchiveRoot(archive)
                aclRoot = acl.create(root)
                return aclRoot.GetRightsByName(self.owner.name)
            except MAPIErrorNotFound:
                pass
        return 0

    def __OpenEntry(self, entry):
        try:
            store = self.storehelper.GetMsgStore(entry[0])
            return (store, store.OpenEntry(entry[1], IID_IMAPIFolder, MAPI_MODIFY))
        except MAPIError:
            return None



class MessageHelper(_BaseHelper):
    def __init__(self, message):
        _BaseHelper.__init__(self, message)
        
        names = [MAPINAMEID(PSETID_Archive, MNID_STRING, u'stubbed')]
        ids = self.object.GetIDsFromNames(names, MAPI_CREATE)
        self.PROP_STUBBED = CHANGE_PROP_TYPE(ids[0], PT_BOOLEAN)

    def IsStubbed(self):
        if not self.IsArchived():
            return False

        try:
            stubbed = HrGetOneProp(self.object, self.PROP_STUBBED)
            return stubbed.Value
        except MAPIError, er:
	    if er.hr != MAPI_E_NOT_FOUND:
		raise
            return False



class ArchiveHelper:
    def __init__(self, object):
        self.object = object;
        
        names = [MAPINAMEID(PSETID_Archive, MNID_STRING, u'ref-store-entryid'),
                 MAPINAMEID(PSETID_Archive, MNID_STRING, u'ref-item-entryid'),
                 MAPINAMEID(PSETID_Archive, MNID_STRING, u'ref-prev-entryid'),
                 MAPINAMEID(PSETID_Archive, MNID_STRING, u'flags')]
        ids = object.GetIDsFromNames(names, MAPI_CREATE)
        self.PROP_REF_STORE_ENTRYID = CHANGE_PROP_TYPE(ids[0], PT_BINARY)
        self.PROP_REF_ITEM_ENTRYID = CHANGE_PROP_TYPE(ids[1], PT_BINARY)
        self.PROP_REF_PREV_ENTRYID = CHANGE_PROP_TYPE(ids[2], PT_BINARY)
        self.PROP_FLAGS = CHANGE_PROP_TYPE(ids[3], PT_LONG)

    def __getattr__(self, attr):
        if attr == 'excludeProps':
            return [self.PROP_REF_STORE_ENTRYID, self.PROP_REF_ITEM_ENTRYID, self.PROP_REF_PREV_ENTRYID, self.PROP_FLAGS]
        raise AttributeError()

    def GetReference(self):
        props = self.object.GetProps([self.PROP_REF_STORE_ENTRYID, self.PROP_REF_ITEM_ENTRYID], 0)
        if props[0].ulPropTag == self.PROP_REF_STORE_ENTRYID or props[1].ulPropTag == self.PROP_REF_ITEM_ENTRYID:
            return (props[0].Value, props[1].Value)
        return (None, None)
