import logging

from helpers import StoreHelper, FolderHelper, MessageHelper, ArchiveHelper

from MAPI import *
from MAPI.Tags import *
from MAPI.Struct import *
from MAPI.Util import *
from MAPI.Util.Generators import GetFolders

def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)

ArchiveOp = enum(NONE = 0, DETACH = 1, UNHOOK = 2, REMOVE = 3)

class ArchiveItem:
    def __init__(self, entryId, archive, lastModTime):
        self.entryId = entryId
        self.archive = archive
        self.lastModTime = lastModTime


class RestoreException(RuntimeError):
    def __init__(self, message):
        Exception.__init__(self, message)

class StoreFull(RestoreException):
    def __init__(self):
        RestoreException.__init__(self, "Quota reached")


def mergeArchives(archives):
    d = {}
    for archive in archives:
        archHelper = ArchiveHelper(archive)
        cTable = archive.GetContentsTable(0)
        cTable.SetColumns([PR_ENTRYID, archHelper.PROP_REF_ITEM_ENTRYID, PR_LAST_MODIFICATION_TIME], TBL_BATCH)
        while True:
            rows = cTable.QueryRows(128, 0)
            if len(rows) == 0:
                break
            for row in rows:
                if PROP_TYPE(row[0].ulPropTag) == PT_ERROR:
                    raise MAPIError(row[0].Value)
                if PROP_TYPE(row[1].ulPropTag) == PT_ERROR:
                    print row[0].Value.encode('hex')
                    raise MAPIError(row[1].Value)
                if not row[1].Value in d or d[row[1].Value].lastModTime < row[2].Value:
                    d[row[1].Value] = ArchiveItem(row[0].Value, archive, row[2].Value)
    e = {}
    for archive in archives:
        e[archive] = [(value.entryId, key) for key, value in d.items() if value.archive is archive]
    return e


def getMsgList(archive):
    l = []
    archHelper = ArchiveHelper(archive)
    cTable = archive.GetContentsTable(0)
    cTable.SetColumns([PR_ENTRYID, archHelper.PROP_REF_ITEM_ENTRYID], TBL_BATCH)
    while True:
        rows = cTable.QueryRows(128, 0)
        if len(rows) == 0:
            break
        for row in rows:
            l.append((row[0].Value, row[1].Value))
    return {archive: l}


def removeAttachments(msg):
    at = msg.GetAttachmentTable(0)
    at.SetColumns([PR_ATTACH_NUM], TBL_BATCH)
    while True:
        rows = at.QueryRows(20, 0)
        if len(rows) == 0:
            break
        for row in rows:
            msg.DeleteAttach(row[0].Value, 0, None, 0)


def restoreMessage(fldHelper, archive, msg):
    global logger
    global dry_run

    # a: archive
    # p: primary
    aId, pId = msg
    session = fldHelper.storehelper.session
    isNew = False

    try:
        pMsg = fldHelper.folder.OpenEntry(pId, IID_IECMessageRaw, MAPI_MODIFY)
        logger.debug("Found primary message.")
    except MAPIError, er:
        if er.hr != MAPI_E_NOT_FOUND:
            raise
        logger.debug("Primary message not found, creating new message")
        pMsg = fldHelper.folder.CreateMessage(None, 0)
        pHelper = MessageHelper(pMsg)
        isNew = True

    pHelper = MessageHelper(pMsg)
    if (not isNew) and (not pHelper.IsStubbed()):
        logger.debug("Primary message is not stubbed.")
        return False # Nothing to do
    removeAttachments(pMsg)
    pMsg.DeleteProps([pHelper.PROP_STUBBED, PR_ICON_INDEX])

    aMsg = archive.OpenEntry(aId, IID_IMessage, MAPI_MODIFY)
    aHelper = ArchiveHelper(aMsg)
    logger.debug("Copying archive message data to primary message.")
    aMsg.CopyTo(None, aHelper.excludeProps, 0, None, IID_IMessage, pMsg, 0)

    try:
        if not dry_run:
            pMsg.SaveChanges(KEEP_OPEN_READONLY)
    except MAPIError, er:
        if er.hr != MAPI_E_STORE_FULL:
            raise
        raise StoreFull()

    if isNew:
        # Update the back ref in the archive so on a subsequent run it will detect
        # the now existing object
        logger.debug("Update reference in archive")
        if not dry_run:
            prop = HrGetOneProp(pMsg, PR_ENTRYID)
            prop.ulPropTag = aHelper.PROP_REF_ITEM_ENTRYID
            aMsg.SetProps([prop])
            aMsg.SaveChanges(0)

    return True


def restoreMessages(fldHelper, archive, msglist):
    global logger

    restored = 0
    skipped = 0
    failed = 0
    try:
        for msg in msglist:
            try:
                if restoreMessage(fldHelper, archive, msg):
                    restored += 1
                else:
                    skipped += 1
            except MAPIError, er:
                logger.exception("Caught unexpected MAPI error, code=0x%08x" % er.hr)
                failed += 1
    finally:
        if restored: logger.info("Restored %u messages" % restored)
        if skipped: logger.info("Skipped %u messages" % skipped)
        if failed: logger.warning("%u messages failed" % failed)


class filterArchives:
    def __init__(self, session, refIds):
        self.session = session
        self.refIds = refIds

    def __call__(self, archive):
        archiveId = HrGetOneProp(archive, PR_STORE_ENTRYID).Value
        for refId in self.refIds:
            if self.session.CompareEntryIDs(archiveId, refId, 0):
                return True
        return False


def restoreFolder(fldHelper, archiveIds):
    global logger

    logger.info("Processing folder '%s'" % HrGetOneProp(fldHelper.folder, PR_DISPLAY_NAME).Value)

    archives = fldHelper.GetArchives()  # returns store/archive tuples.
    logger.debug("Found %u archives for folder" % len(archives))

    archives = [item[1] for item in archives if item is not None]   # Get the archives if not None.
    logger.debug("%u archives are valid" % len(archives))

    if archiveIds is not None and len(archiveIds) > 0:
        archives = filter(filterArchives(fldHelper.storehelper.session, archiveIds), archives)
    logger.debug("Got %u archives after filtering" % len(archives))
    if len(archives) == 0:
        logger.debug("No archives to process")
        return
    if len(archives) > 1:
        msglists = mergeArchives(archives)
        logger.debug("Merged archives into %u set(s)" % len(msglists))
    else:
        msglists = getMsgList(archives[0])
    for archive, msglist in msglists.items():
        logger.debug("Processing %u message(s) from set" % len(msglist))
        restoreMessages(fldHelper, archive, msglist)


def UnhookOrRemoveArchive(store, op):
    guid, userId = [prop.Value for prop in store.GetProps([PR_STORE_RECORD_KEY, PR_MAILBOX_OWNER_ENTRYID], 0)]
    sa = store.QueryInterface(IID_IECServiceAdmin)
    if op >= ArchiveOp.UNHOOK:
        sa.UnhookStore(ECSTORE_TYPE_ARCHIVE, userId)
        if op == ArchiveOp.REMOVE:
            logger.debug("Store GUID = '%s'" % guid.encode('hex'))
            sa.RemoveStore(guid)


def main(session, user, **kwargs):
    global logger
    global dry_run

    if 'logger' in kwargs:
        logger = kwargs['logger']
    else:
        logger = logging.getLogger('')
    if 'dry_run' in kwargs:
        dry_run = kwargs['dry_run']
    else:
        dry_run = False
    if 'archive_op' in kwargs:
        archive_op = kwargs['archive_op']
    else:
        archive_op = ArchiveOp.NONE

    if 'archiveIds' in kwargs:
        archiveIds = kwargs['archiveIds']
    else:
        archiveIds = None
    if archiveIds is not None and len(archiveIds) == 0:
        logger.fatal("No archives specified. Nothing to restore")
        return 0

    ems = GetDefaultStore(session).QueryInterface(IID_IExchangeManageStore)
    storeId = ems.CreateStoreEntryID(None, user, 0)
    store = session.OpenMsgStore(0, storeId, None, MDB_WRITE)
    storeHelper = StoreHelper(session, store)

    if archiveIds is not None:
        logger.debug("Got %u archiveIds from caller" % len(archiveIds))
        tempIds = []
        archives = storeHelper.GetArchiveStores()
        logger.debug("Got %u archives from store" % len(archiveIds))
        for archiveId in archiveIds:
            if archiveId >= len(archives):
                logger.fatal("Archive index out of range: %u. Valid range: 0 - %u" % (archiveId, len(archives) - 1))
                return 1
            tempIds.append(HrGetOneProp(archives[archiveId], PR_ENTRYID).Value)
        archiveIds = tempIds
        logger.debug("Got %u archiveIds after combining" % len(archiveIds))

    # We can now safely remove the selected or all archives from the list
    # of attached archives on the store. Effectively detaching it.
    removableStores = []
    if archive_op >= ArchiveOp.DETACH:
        logger.info("Detaching archives")
        if not dry_run:
            removableStores = storeHelper.DetachArchives(archiveIds)  # Passing None detaches all

    ipmSubtreeId = HrGetOneProp(store, PR_IPM_SUBTREE_ENTRYID).Value
    try:
        for folder, depth in GetFolders(store, rootid=ipmSubtreeId):
            fldHelper = FolderHelper(storeHelper, folder)
            restoreFolder(fldHelper, archiveIds)
    except RestoreException, re:
        logger.fatal("Operation cancelled: %s" % re.message)
        return 1

    if (archive_op >= ArchiveOp.UNHOOK):
        if archive_op == ArchiveOp.UNHOOK:
            action = "Unhooking"
        else:
            action = "Removing"
        logger.info("%s %u archive stores" % (action, len(removableStores)))
        if not dry_run:
            for store in removableStores:
                UnhookOrRemoveArchive(store, archive_op)
